@@ -7,8 +7,7 @@
 #      PODNAME:  Games::Go::AGA::DataObjects::Register
 #     ABSTRACT:  models AGA register.tde file information
 #
-#       AUTHOR:  Reid Augustin (REID), <reid@lucidport.com>
-#      COMPANY:  LucidPort Technology, Inc.
+#       AUTHOR:  Reid Augustin (REID), <reid@hellosix.com>
 #      CREATED:  11/19/2010 03:13:05 PM PST
 #===============================================================================
 
@@ -17,8 +16,9 @@ use strict;
 use warnings;
 
 package Games::Go::AGA::DataObjects::Register;
+use Moo;
+use namespace::clean;
 use parent 'Games::Go::AGA::DataObjects::Directives';
-use Mouse;
 
 use Carp;
 use Readonly;
@@ -28,8 +28,9 @@ use Games::Go::AGA::Parse::Util qw( normalize_ID );
 use Games::Go::AGA::Parse::Util qw( Rank_to_Rating );
 use Games::Go::AGA::DataObjects::Player;
 use Games::Go::AGA::DataObjects::Directives;
+use Games::Go::AGA::DataObjects::Types qw( isa_CodeRef isa_ArrayRef isa_HashRef );
 
-our $VERSION = '0.107'; # VERSION
+our $VERSION = '0.152'; # VERSION
 
 #   has 'directives' => (
 #       isa => 'Games::Go::AGA::DataObjects::Directives',
@@ -48,22 +49,142 @@ our $VERSION = '0.107'; # VERSION
 #       is => 'ro',
 #       default => sub { [] }
 #   );
-has 'change_callback' => (
-    isa => 'Maybe[CodeRef]',
+has change_callback => (
+    isa => \&isa_CodeRef,
     is => 'rw',
+    lazy => 1,
     default => sub { sub { } }
 );
-
+has _bye_candidates => (
+    is => 'rw',
+);
+has _drops => (
+    is => 'rw',
+);
 sub BUILD {
     my ($self) = @_;
     $self->{comments} = [];
     $self->{players} = [];
 }
 
+sub directive_is_boolean { return $_[0]->is_boolean($_[1]) }    # proxy to Directives
+
+# hashref of player IDs, each value true or false
+sub bye_candidates {
+    my ($self) = @_;
+
+    if (not $self->_bye_candidates) {
+        my %bye_candidates = map { $_ => $self->get_player($_) }
+            split /\s+/, $self->get_directive_value('BYE_CANDIDATES') || '';
+        $self->_bye_candidates(\%bye_candidates);
+    }
+    return $self->_bye_candidates;
+}
+
+sub add_bye_candidates {
+    my ($self, @candidates) = @_;
+
+    my $bye_candidates_hash = $self->bye_candidates;
+
+    for my $candidate (@candidates) {
+        if (not $bye_candidates_hash->{$candidate}) {
+            $self->_find_player_idx($candidate);   # croaks if not found
+            $self->_bye_candidates(undef);  # force refresh on next ->bye_candidates
+            $bye_candidates_hash->{$candidate} = 1;
+        }
+    }
+    if (not $self->_bye_candidates) {
+        $self->set_directive_value('BYE_CANDIDATES', join(' ', keys %{$bye_candidates_hash}));
+    }
+}
+
+sub delete_bye_candidates {
+    my ($self, @candidates) = @_;
+
+    my $bye_candidates_hash = $self->bye_candidates;
+
+    my @deleted;
+    for my $candidate (@candidates) {
+        if ($bye_candidates_hash->{$candidate}) {
+            $self->_bye_candidates(undef);  # force refresh on next ->bye_candidates
+            push @deleted, $candidate if (delete $bye_candidates_hash->{$candidate});
+        }
+    }
+    if (not $self->_bye_candidates) {
+        $self->set_directive_value('BYE_CANDIDATES', join(' ', keys %{$bye_candidates_hash}));
+    }
+    return @deleted;
+}
+
+# arrayref of rounds, each entry a hashref of player IDs, true or false
+sub drops {
+    my ($self, $round_num) = @_;
+
+    if (not $self->_drops) {
+        my @drops;
+        for my $ii (0 .. ($self->get_directive_value('ROUNDS') || 7)) {
+            my $key = $ii ? "DROP_$ii" : 'DROP_ALL';
+            my %round_drops = map { $_ => $self->get_player($_) }
+                split /\s+/, $self->get_directive_value($key) || '';
+            push @drops, \%round_drops;
+        }
+        $self->_drops(\@drops);
+    }
+    return $self->_drops->[$round_num] if (@_ > 1);
+    return $self->_drops;
+}
+
+sub add_drops {
+    my ($self, $round_num, @ids) = @_;
+
+    my $drops_hash = $self->drops($round_num);
+
+    for my $id (@ids) {
+        if (not $drops_hash->{$id}) {
+            $self->_find_player_idx($id);   # croaks if not found
+            $self->_drops(undef);   # force refresh on next ->drops
+            $drops_hash->{$id} = 1;
+        }
+    }
+    if (not $self->_drops) {
+        $round_num ||= 'All';   # round 0 means All
+        $self->set_directive_value("DROP_$round_num", join(' ', keys %{$drops_hash}));
+    }
+}
+
+sub delete_drops {
+    my ($self, $round_num, @ids) = @_;
+
+    if (not defined $round_num) {   # all rounds
+        my @dropped_ary;
+        my $ii = 0;
+        for my $drop_round (@{$self->drops}) {
+            $dropped_ary[$ii] = { map { $_ => 1 } $self->delete_drops($ii, @ids) } || {};
+            $ii++;
+        }
+        return @dropped_ary;
+    }
+
+    my $drops_hash = $self->drops($round_num);
+
+    my @deleted;
+    for my $id (@ids) {
+        if ($drops_hash->{$id}) {
+            $self->_drops(undef);   # force refresh on next ->drops
+            push @deleted, $id if (delete $drops_hash->{$id});
+        }
+    }
+    if (not $self->_drops) {
+        $round_num ||= 'All';   # round 0 means All
+        $self->set_directive_value("DROP_$round_num", join(' ', keys %{$drops_hash}));
+    }
+    return @deleted
+}
+
 sub changed {
     my ($self) = @_;
 
-    &{$self->change_callback}($self) if ($self->{change_callback});
+    &{$self->change_callback}(@_);
 }
 
 sub comments {
@@ -99,10 +220,11 @@ sub id_is_duplicate {
     return 0;
 }
 
+sub add_player { shift->insert_player_at_idx(-1, @_) }
+
 sub insert_player_at_idx {
     my ($self, $idx, $player) = @_;
 
-
     my $id = $player->id;
     if (not $id) {
         $id = 1;
@@ -148,6 +270,8 @@ sub delete_player {
         croak "Games recorded for $id, can't delete\n";
     }
     my $player = splice(@{$players}, $idx, 1);   # delete and return it
+    $self->delete_bye_candidates($player->id);
+    $self->delete_drops(undef, $player->id);
 #print "delete ", $player->id, "\n";
     $self->changed;
     return $player;
@@ -178,7 +302,7 @@ sub _find_player_idx {
     else {
         # find Player with matching ID
         FIND_ID : {
-            my $id = $idx;
+            my $id = normalize_ID($idx);
             for my $ii (0 .. $#{$players}) {
                 if ($players->[$ii]->id eq $id) {
                     $idx = $ii;
@@ -196,7 +320,26 @@ sub _find_player_idx {
 }
 
 # override some Directives methods so we can intercept
-# BAND_BREAKS directives (which requires player info)
+# setting/getting of certain directives (which requires player info)
+sub set_directive_value {
+    my $self = shift;
+    my ($key, $val) = @_;
+
+    if (uc $key eq 'BYE_CANDIDATES') {
+        $self->_bye_candidates(undef);  # force refresh on next ->bye_candidates
+        if (not $val) {
+            return $self->SUPER::delete_directive(@_);
+        }
+    }
+    elsif ($key =~ m/^DROP_(\d+|ALL)$/i) {
+        $self->_drops(undef);  # force refresh on next ->drops
+        if (not $val) {
+            return $self->SUPER::delete_directive(@_);
+        }
+    }
+    return $self->SUPER::set_directive_value(@_);
+}
+
 sub get_directive_value {
     my $self = shift;
     my ($key) = $_[0];
@@ -228,6 +371,7 @@ sub break_bands {
     my $ii = 0;
     my @band_breaks;
     for my $band (1 .. $num_bands - 1) {
+        last if ($ii >= @sorted_ranks);
         my $next_break = $band * (@players / $num_bands);
         while ($running_total + $entrants_per_rating{$sorted_ranks[$ii]} < $next_break) {
             $running_total += $entrants_per_rating{$sorted_ranks[$ii++]};
@@ -237,7 +381,6 @@ sub break_bands {
         }
 #print "$running_total to band $band=$sorted_ranks[$ii - 1]\n";
         push(@band_breaks, $sorted_ranks[$ii - 1]);
-        last if ($ii >= @sorted_ranks);
     }
     $band_breaks = join ' ', @band_breaks;
     $self->set_directive_value('BAND_BREAKS', $band_breaks);
@@ -287,9 +430,6 @@ sub fprint {
     }
 }
 
-no Mouse;
-__PACKAGE__->meta->make_immutable;
-
 1;
 
 __END__
@@ -304,7 +444,7 @@ Games::Go::AGA::DataObjects::Register - models AGA register.tde file information
 
 =head1 VERSION
 
-version 0.107
+version 0.152
 
 =head1 SYNOPSIS
 
@@ -516,7 +656,7 @@ Reid Augustin <reid@hellosix.com>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2015 by Reid Augustin.
+This software is copyright (c) 2016 by Reid Augustin.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
